from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

import attr

if TYPE_CHECKING:
    from ..models.active import Active
    from ..models.ape import Ape
    from ..models.ape_data import ApeData
    from ..models.inactive import Inactive
    from ..models.plugin_external_data import PluginExternalData


T = TypeVar("T", bound="Plugin")


@attr.s(auto_attribs=True)
class Plugin:
    """
    Attributes:
        data (ApeData):
        external_data (PluginExternalData):
        name (Ape):
        status (Union['Active', 'Inactive']):
    """

    data: "ApeData"
    external_data: "PluginExternalData"
    name: "Ape"
    status: Union["Active", "Inactive"]
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.active import Active

        data = self.data.to_dict()

        external_data = self.external_data.to_dict()

        name = self.name.to_dict()

        status: Dict[str, Any]

        if isinstance(self.status, Active):
            status = self.status.to_dict()

        else:
            status = self.status.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "data": data,
                "externalData": external_data,
                "name": name,
                "status": status,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.active import Active
        from ..models.ape import Ape
        from ..models.ape_data import ApeData
        from ..models.inactive import Inactive
        from ..models.plugin_external_data import PluginExternalData

        d = src_dict.copy()
        data = ApeData.from_dict(d.pop("data"))

        external_data = PluginExternalData.from_dict(d.pop("externalData"))

        name = Ape.from_dict(d.pop("name"))

        def _parse_status(data: object) -> Union["Active", "Inactive"]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                componentsschemas_plugin_status_type_0 = Active.from_dict(data)

                return componentsschemas_plugin_status_type_0
            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            componentsschemas_plugin_status_type_1 = Inactive.from_dict(data)

            return componentsschemas_plugin_status_type_1

        status = _parse_status(d.pop("status"))

        plugin = cls(
            data=data,
            external_data=external_data,
            name=name,
            status=status,
        )

        plugin.additional_properties = d
        return plugin

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from .auth import AuthState
from priceloop_api import Client
from typing import Dict
import attr

default_host = "alpha.priceloop.ai"


class PriceloopClient(Client):
    """A Client which has been authenticated for use on secured endpoints"""

    auth_state: AuthState

    def __init__(self, auth_state: AuthState):
        self.base_url = auth_state.config().base_url
        self.timeout = 30.0
        self.cookies = {}
        self.headers = {}
        self.verify_ssl = True
        self.raise_on_unexpected_status = False
        self.auth_state = auth_state

    def get_headers(self) -> Dict[str, str]:
        """Get headers to be used in authenticated endpoints"""
        auth_header_value = f"Bearer {self.auth_state.access_token()}"
        return {"Authorization": auth_header_value, **self.headers}

    @staticmethod
    def with_credentials(username: str, password: str, host: str = default_host):
        auth_state = AuthState(username, password, host)
        return PriceloopClient(auth_state)



from enum import Enum


class AddDataColumnType(str, Enum):
    STRING = "string"
    NUMBER = "number"
    BOOLEAN = "boolean"
    DATE = "date"
    JSON = "json"

    def __str__(self) -> str:
        return str(self.value)



from dataclasses import dataclass
from datetime import datetime, timedelta

import boto3
import requests


@dataclass(frozen=True)
class ApiConfig:
    region: str
    client_id: str
    api_scopes: str
    auth_url: str
    base_url: str


@dataclass(frozen=True)
class AuthTokens:
    access_token: str
    expires_in_seconds: int
    expiry_date: datetime
    id_token: str
    refresh_token: str
    token_type: str


class AuthState(object):
    nocode_config: ApiConfig = None
    auth_tokens: AuthTokens = None

    def config(self):
        if self.nocode_config is None:
            endpoint = f"https://{self.host}/app_config.json"

            self.nocode_config = get_config(endpoint)

        return self.nocode_config


    def access_token(self):
        if self.auth_tokens is None or is_expired(datetime.now(), self.auth_tokens):
            self.auth_tokens = get_oauth_tokens(self.config(), self.username, self.password)

        return self.auth_tokens.access_token

    def __init__(self, username, password, host):
        self.username = username
        self.password = password
        self.host = host


def is_expired(now: datetime, tokens: AuthTokens) -> bool:
    return tokens.expiry_date - timedelta(minutes=1) > now


def get_config(app_config_url: str, auth_endpoint=None) -> ApiConfig:
    response = requests.get(app_config_url)
    cfg = response.json()
    endpoint = None

    if auth_endpoint is None:
        endpoint = cfg["auth"]["url"]
    else:
        endpoint = auth_endpoint

    return ApiConfig(
        region=cfg["region"],
        client_id=cfg["auth"]["clientId"],
        api_scopes=cfg["auth"]["apiScopes"],
        auth_url=endpoint,
        base_url=cfg["http"]["url"],
    )


def get_oauth_tokens(config: ApiConfig, user_email: str, user_password: str) -> AuthTokens:
    cognito_client = boto3.client("cognito-idp", region_name=config.region)

    auth_response = cognito_client.initiate_auth(
        AuthFlow="USER_PASSWORD_AUTH",
        AuthParameters={"USERNAME": user_email, "PASSWORD": user_password},
        ClientId=config.client_id,
    )

    # calculate expiry date, because cognito only gives the number of seconds
    now = datetime.today()
    expires_in_seconds = int(auth_response["AuthenticationResult"]["ExpiresIn"])
    expiry_date = now + timedelta(seconds=expires_in_seconds)

    return AuthTokens(
        access_token=auth_response["AuthenticationResult"]["AccessToken"],
        expires_in_seconds=expires_in_seconds,
        expiry_date=expiry_date,
        id_token=auth_response["AuthenticationResult"]["IdToken"],
        refresh_token=auth_response["AuthenticationResult"]["RefreshToken"],
        token_type=auth_response["AuthenticationResult"]["TokenType"],
    )



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

import attr
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.csv_separator import CsvSeparator
    from ..models.s3_key import S3Key


T = TypeVar("T", bound="ImportJob")


@attr.s(auto_attribs=True)
class ImportJob:
    """
    Attributes:
        created_at (datetime.datetime):
        csv_separator (CsvSeparator):
        id (int):
        s_3_key (S3Key):
        table_import_mode (str):
        table_name (str):
        finished_at (Union[Unset, datetime.datetime]):
        is_successful (Union[Unset, bool]):
        message (Union[Unset, str]):
        started_at (Union[Unset, datetime.datetime]):
    """

    created_at: datetime.datetime
    csv_separator: "CsvSeparator"
    id: int
    s_3_key: "S3Key"
    table_import_mode: str
    table_name: str
    finished_at: Union[Unset, datetime.datetime] = UNSET
    is_successful: Union[Unset, bool] = UNSET
    message: Union[Unset, str] = UNSET
    started_at: Union[Unset, datetime.datetime] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        created_at = self.created_at.isoformat()

        csv_separator = self.csv_separator.to_dict()

        id = self.id
        s_3_key = self.s_3_key.to_dict()

        table_import_mode = self.table_import_mode
        table_name = self.table_name
        finished_at: Union[Unset, str] = UNSET
        if not isinstance(self.finished_at, Unset):
            finished_at = self.finished_at.isoformat()

        is_successful = self.is_successful
        message = self.message
        started_at: Union[Unset, str] = UNSET
        if not isinstance(self.started_at, Unset):
            started_at = self.started_at.isoformat()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "createdAt": created_at,
                "csvSeparator": csv_separator,
                "id": id,
                "s3Key": s_3_key,
                "tableImportMode": table_import_mode,
                "tableName": table_name,
            }
        )
        if finished_at is not UNSET:
            field_dict["finishedAt"] = finished_at
        if is_successful is not UNSET:
            field_dict["isSuccessful"] = is_successful
        if message is not UNSET:
            field_dict["message"] = message
        if started_at is not UNSET:
            field_dict["startedAt"] = started_at

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.csv_separator import CsvSeparator
        from ..models.s3_key import S3Key

        d = src_dict.copy()
        created_at = isoparse(d.pop("createdAt"))

        csv_separator = CsvSeparator.from_dict(d.pop("csvSeparator"))

        id = d.pop("id")

        s_3_key = S3Key.from_dict(d.pop("s3Key"))

        table_import_mode = d.pop("tableImportMode")

        table_name = d.pop("tableName")

        _finished_at = d.pop("finishedAt", UNSET)
        finished_at: Union[Unset, datetime.datetime]
        if isinstance(_finished_at, Unset):
            finished_at = UNSET
        else:
            finished_at = isoparse(_finished_at)

        is_successful = d.pop("isSuccessful", UNSET)

        message = d.pop("message", UNSET)

        _started_at = d.pop("startedAt", UNSET)
        started_at: Union[Unset, datetime.datetime]
        if isinstance(_started_at, Unset):
            started_at = UNSET
        else:
            started_at = isoparse(_started_at)

        import_job = cls(
            created_at=created_at,
            csv_separator=csv_separator,
            id=id,
            s_3_key=s_3_key,
            table_import_mode=table_import_mode,
            table_name=table_name,
            finished_at=finished_at,
            is_successful=is_successful,
            message=message,
            started_at=started_at,
        )

        import_job.additional_properties = d
        return import_job

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



""" Contains all the data models used in inputs/outputs """

from .active import Active
from .add_data_column_type import AddDataColumnType
from .amazon import Amazon
from .ape import Ape
from .ape_data import ApeData
from .ape_data_typeform import ApeDataTypeform
from .api_column import ApiColumn
from .api_external_function import ApiExternalFunction
from .api_table import ApiTable
from .api_table_data import ApiTableData
from .api_workspace import ApiWorkspace
from .create_csv_import_job_mode import CreateCsvImportJobMode
from .create_external_function_return_type import CreateExternalFunctionReturnType
from .create_external_function_runtime import CreateExternalFunctionRuntime
from .csv_separator import CsvSeparator
from .get_table_upload_csv_url_mode import GetTableUploadCsvUrlMode
from .import_job import ImportJob
from .import_job_response import ImportJobResponse
from .inactive import Inactive
from .plugin import Plugin
from .plugin_external_data import PluginExternalData
from .plugin_tokens import PluginTokens
from .presigned_url import PresignedUrl
from .s3_key import S3Key
from .set_plugin_external_data_json_body import SetPluginExternalDataJsonBody
from .set_plugin_status_status import SetPluginStatusStatus
from .table_row import TableRow

__all__ = (
    "Active",
    "AddDataColumnType",
    "Amazon",
    "Ape",
    "ApeData",
    "ApeDataTypeform",
    "ApiColumn",
    "ApiExternalFunction",
    "ApiTable",
    "ApiTableData",
    "ApiWorkspace",
    "CreateCsvImportJobMode",
    "CreateExternalFunctionReturnType",
    "CreateExternalFunctionRuntime",
    "CsvSeparator",
    "GetTableUploadCsvUrlMode",
    "ImportJob",
    "ImportJobResponse",
    "Inactive",
    "Plugin",
    "PluginExternalData",
    "PluginTokens",
    "PresignedUrl",
    "S3Key",
    "SetPluginExternalDataJsonBody",
    "SetPluginStatusStatus",
    "TableRow",
)



from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="TableRow")


@attr.s(auto_attribs=True)
class TableRow:
    """
    Attributes:
        id (int):
        values (Union[Unset, List[str]]):
    """

    id: int
    values: Union[Unset, List[str]] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        values: Union[Unset, List[str]] = UNSET
        if not isinstance(self.values, Unset):
            values = self.values

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
            }
        )
        if values is not UNSET:
            field_dict["values"] = values

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        values = cast(List[str], d.pop("values", UNSET))

        table_row = cls(
            id=id,
            values=values,
        )

        table_row.additional_properties = d
        return table_row

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import requests
import pandas as pd
from io import StringIO

from priceloop_api.models.create_csv_import_job_mode import CreateCsvImportJobMode
from priceloop_api.client import AuthenticatedClient
from priceloop_api.api.default import (
    list_workspaces,
    get_workspace,
    get_table,
    get_table_data,
    create_csv_import_job,
)

d = {"t": True, "f": False}


def to_nocode(
    df: pd.DataFrame,
    table_name: str,
    client: AuthenticatedClient,
    mode: CreateCsvImportJobMode = CreateCsvImportJobMode.DELETE_AND_RECREATE,
    workspace_name: str = None,
) -> None:
    csv_buffer = StringIO()
    df.to_csv(csv_buffer, index=None)

    if workspace_name is None:
        workspaces = list_workspaces.sync(client=client)
        workspace = get_workspace.sync(workspaces[0], client=client)
        workspace_name = workspace.name

    import_job = create_csv_import_job.sync(workspace_name, table_name, mode=mode, client=client)

    requests.put(import_job.put_url, data=csv_buffer.getvalue().encode("utf-8"))
    print("Upload Successful, please wait a moment for the changes to appear")


def read_nocode(
    table_name: str, client: AuthenticatedClient, limit: int, offset: int, workspace_name: str = None
) -> pd.DataFrame:
    csv_buffer = StringIO()

    if workspace_name is None:
        workspaces = list_workspaces.sync(client=client)
        workspace = get_workspace.sync(workspaces[0], client=client)

        workspace_name = workspace.name

    raw_header = get_table.sync(workspace_name, table_name, client=client)

    header = [i.name for i in raw_header.columns]
    boolean_cols = [i.name for i in raw_header.columns if i.tpe == "boolean"]
    raw_table_data = get_table_data.sync(workspace_name, table_name, client=client, limit=limit, offset=offset)
    table_data = pd.DataFrame([v.values for v in raw_table_data.rows], columns=header)
    for col in boolean_cols:
        table_data[col] = table_data[col].map(d)
    # To-do: infer type from nocode
    table_data.to_csv(csv_buffer, index=None)
    csv_buffer.seek(0)
    table_data_type_inferred = pd.read_csv(csv_buffer)
    return table_data_type_inferred



from enum import Enum


class GetTableUploadCsvUrlMode(str, Enum):
    NEW = "new"
    DELETE_AND_RECREATE = "delete_and_recreate"
    REPLACE_DATA = "replace_data"
    APPEND_DATA = "append_data"

    def __str__(self) -> str:
        return str(self.value)



from http import HTTPStatus
from typing import Any, Dict, List, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.create_external_function_return_type import CreateExternalFunctionReturnType
from ...models.create_external_function_runtime import CreateExternalFunctionRuntime
from ...models.presigned_url import PresignedUrl
from ...types import UNSET, Response, Unset


def _get_kwargs(
    workspace: str,
    function: str,
    *,
    client: AuthenticatedClient,
    runtime: CreateExternalFunctionRuntime,
    return_type: CreateExternalFunctionReturnType,
    param_type: Union[Unset, None, List[str]] = UNSET,
) -> Dict[str, Any]:
    url = "{}/api/v1.0/workspaces/{workspace}/external-functions/{function}".format(
        client.base_url, workspace=workspace, function=function
    )

    headers: Dict[str, str] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    params: Dict[str, Any] = {}
    json_runtime = runtime.value

    params["runtime"] = json_runtime

    json_return_type = return_type.value

    params["returnType"] = json_return_type

    json_param_type: Union[Unset, None, List[str]] = UNSET
    if not isinstance(param_type, Unset):
        if param_type is None:
            json_param_type = None
        else:
            json_param_type = param_type

    params["paramType"] = json_param_type

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    return {
        "method": "post",
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
        "params": params,
    }


def _parse_response(*, client: Client, response: httpx.Response) -> Optional[PresignedUrl]:
    if response.status_code == HTTPStatus.OK:
        response_200 = PresignedUrl.from_dict(response.json())

        return response_200
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(f"Unexpected status code: {response.status_code}")
    else:
        return None


def _build_response(*, client: Client, response: httpx.Response) -> Response[PresignedUrl]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    workspace: str,
    function: str,
    *,
    client: AuthenticatedClient,
    runtime: CreateExternalFunctionRuntime,
    return_type: CreateExternalFunctionReturnType,
    param_type: Union[Unset, None, List[str]] = UNSET,
) -> Response[PresignedUrl]:
    """Create a new external function, specifying the function name, runtime, return type and the paremeter
    types. This API endpoint returns a url, to which you can upload your function code. You can do a PUT
    request on the returned url, e.g.: curl -XPUT -T <zip-file> '<url>'.

    Args:
        workspace (str):
        function (str):
        runtime (CreateExternalFunctionRuntime):
        return_type (CreateExternalFunctionReturnType):
        param_type (Union[Unset, None, List[str]]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[PresignedUrl]
    """

    kwargs = _get_kwargs(
        workspace=workspace,
        function=function,
        client=client,
        runtime=runtime,
        return_type=return_type,
        param_type=param_type,
    )

    response = httpx.request(
        verify=client.verify_ssl,
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    workspace: str,
    function: str,
    *,
    client: AuthenticatedClient,
    runtime: CreateExternalFunctionRuntime,
    return_type: CreateExternalFunctionReturnType,
    param_type: Union[Unset, None, List[str]] = UNSET,
) -> Optional[PresignedUrl]:
    """Create a new external function, specifying the function name, runtime, return type and the paremeter
    types. This API endpoint returns a url, to which you can upload your function code. You can do a PUT
    request on the returned url, e.g.: curl -XPUT -T <zip-file> '<url>'.

    Args:
        workspace (str):
        function (str):
        runtime (CreateExternalFunctionRuntime):
        return_type (CreateExternalFunctionReturnType):
        param_type (Union[Unset, None, List[str]]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[PresignedUrl]
    """

    return sync_detailed(
        workspace=workspace,
        function=function,
        client=client,
        runtime=runtime,
        return_type=return_type,
        param_type=param_type,
    ).parsed


async def asyncio_detailed(
    workspace: str,
    function: str,
    *,
    client: AuthenticatedClient,
    runtime: CreateExternalFunctionRuntime,
    return_type: CreateExternalFunctionReturnType,
    param_type: Union[Unset, None, List[str]] = UNSET,
) -> Response[PresignedUrl]:
    """Create a new external function, specifying the function name, runtime, return type and the paremeter
    types. This API endpoint returns a url, to which you can upload your function code. You can do a PUT
    request on the returned url, e.g.: curl -XPUT -T <zip-file> '<url>'.

    Args:
        workspace (str):
        function (str):
        runtime (CreateExternalFunctionRuntime):
        return_type (CreateExternalFunctionReturnType):
        param_type (Union[Unset, None, List[str]]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[PresignedUrl]
    """

    kwargs = _get_kwargs(
        workspace=workspace,
        function=function,
        client=client,
        runtime=runtime,
        return_type=return_type,
        param_type=param_type,
    )

    async with httpx.AsyncClient(verify=client.verify_ssl) as _client:
        response = await _client.request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    workspace: str,
    function: str,
    *,
    client: AuthenticatedClient,
    runtime: CreateExternalFunctionRuntime,
    return_type: CreateExternalFunctionReturnType,
    param_type: Union[Unset, None, List[str]] = UNSET,
) -> Optional[PresignedUrl]:
    """Create a new external function, specifying the function name, runtime, return type and the paremeter
    types. This API endpoint returns a url, to which you can upload your function code. You can do a PUT
    request on the returned url, e.g.: curl -XPUT -T <zip-file> '<url>'.

    Args:
        workspace (str):
        function (str):
        runtime (CreateExternalFunctionRuntime):
        return_type (CreateExternalFunctionReturnType):
        param_type (Union[Unset, None, List[str]]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[PresignedUrl]
    """

    return (
        await asyncio_detailed(
            workspace=workspace,
            function=function,
            client=client,
            runtime=runtime,
            return_type=return_type,
            param_type=param_type,
        )
    ).parsed



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.table_row import TableRow


T = TypeVar("T", bound="ApiTableData")


@attr.s(auto_attribs=True)
class ApiTableData:
    """
    Attributes:
        rows (Union[Unset, List['TableRow']]):
    """

    rows: Union[Unset, List["TableRow"]] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        rows: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.rows, Unset):
            rows = []
            for rows_item_data in self.rows:
                rows_item = rows_item_data.to_dict()

                rows.append(rows_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if rows is not UNSET:
            field_dict["rows"] = rows

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.table_row import TableRow

        d = src_dict.copy()
        rows = []
        _rows = d.pop("rows", UNSET)
        for rows_item_data in _rows or []:
            rows_item = TableRow.from_dict(rows_item_data)

            rows.append(rows_item)

        api_table_data = cls(
            rows=rows,
        )

        api_table_data.additional_properties = d
        return api_table_data

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.amazon import Amazon


T = TypeVar("T", bound="PluginTokens")


@attr.s(auto_attribs=True)
class PluginTokens:
    """
    Attributes:
        amazon (Union[Unset, Amazon]):
    """

    amazon: Union[Unset, "Amazon"] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        amazon: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.amazon, Unset):
            amazon = self.amazon.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if amazon is not UNSET:
            field_dict["amazon"] = amazon

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.amazon import Amazon

        d = src_dict.copy()
        _amazon = d.pop("amazon", UNSET)
        amazon: Union[Unset, Amazon]
        if isinstance(_amazon, Unset):
            amazon = UNSET
        else:
            amazon = Amazon.from_dict(_amazon)

        plugin_tokens = cls(
            amazon=amazon,
        )

        plugin_tokens.additional_properties = d
        return plugin_tokens

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

import attr

if TYPE_CHECKING:
    from ..models.ape_data_typeform import ApeDataTypeform


T = TypeVar("T", bound="ApeData")


@attr.s(auto_attribs=True)
class ApeData:
    """
    Attributes:
        typeform (ApeDataTypeform):
    """

    typeform: "ApeDataTypeform"
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        typeform = self.typeform.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "typeform": typeform,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.ape_data_typeform import ApeDataTypeform

        d = src_dict.copy()
        typeform = ApeDataTypeform.from_dict(d.pop("typeform"))

        ape_data = cls(
            typeform=typeform,
        )

        ape_data.additional_properties = d
        return ape_data

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
